import numpy as np


W = 4172  # 4 * 7 * 149
H = 1202  # 2 * 601


def cut(np_img, n_s_row_max, n_s_col_max, overlap_w=-1, overlap_h=-1):
    """
    :return the image cut in pieces and the number of images per column.
            The pieces are overlapping, you can set how much with the overlap_w and overlap_h params.
    :param np_img: image as numpy array
    :param n_s_row_max: number of rows expected in each single image returned (it can change in the edge cases)
    :param n_s_col_max: number of columns expected in each single image returned (it can change in the edge cases)
    :param overlap_w: how much the images overlap with each other on the x axis. (it is guaranteed to be constant)
                if it's -1 it is set to half the width of the img
    :param overlap_h: how much the images overlap with each other on the y axis. (it is guaranteed to be constant)
                if it's -1 it is set to half the height of the img
    """
    W = np_img.shape[1]
    H = np_img.shape[0]
    def_overlaps = get_default_overlaps(n_s_row_max, n_s_col_max)
    if overlap_h == -1:
        overlap_h = def_overlaps[0]
    if overlap_w == -1:
        overlap_w = def_overlaps[1]
    n_img_per_col = 0
    # pad tells how much should we move when changing index while looping
    pad_w = n_s_col_max - overlap_w
    pad_h = n_s_row_max - overlap_h
    if overlap_w >= n_s_col_max or overlap_h >= n_s_row_max:
        raise ValueError("overlap is major than the size of the cut img")
    h_rem = get_rem(H, n_s_row_max, pad_h)
    w_rem = get_rem(W, n_s_col_max, pad_w)

    ret_img = []
    # dim_img contains the width and height of each img. it can be useful for reconstruction.
    for i in get_range(H, n_s_row_max, pad_h):
        n_img_per_col += 1
        slice_rows = get_slice(i, n_s_row_max, pad_h)
        for j in get_range(W, n_s_col_max, pad_w):
            slice_cols = get_slice(j, n_s_col_max, pad_w)
            ret_img.append(np_img[slice_rows, slice_cols, :])
        # handle remaining w
        if not (w_rem == 0):
            # There is some w left
            # we have to be sure that overlaps overlap_w
            slice_rows = get_slice(i, n_s_row_max, pad_h)
            ret_img.append(np_img[slice_rows, get_edge_start_slice(w_rem, overlap_w):, :])
    # handle remaining h
    if not (h_rem == 0):
        # There is some h left
        # we have to be sure that overlaps overlap_h
        n_img_per_col += 1
        h_start_slice = get_edge_start_slice(h_rem, overlap_h)
        for j in get_range(W, n_s_col_max, pad_w):
            slice_cols = get_slice(j, n_s_col_max, pad_w)
            ret_img.append(np_img[h_start_slice:, slice_cols, :])
        if not (w_rem == 0):
            # There is some w left
            # we have to be sure that overlaps overlap_w
            ret_img.append(np_img[h_start_slice:, get_edge_start_slice(w_rem, overlap_w):, :])
    return [ret_img, n_img_per_col]


def get_rem(dim, n_s_dim, pad):
    rem = 0
    if ((dim - n_s_dim) // pad) * pad != dim - n_s_dim:
        rem = (dim - n_s_dim) % pad
    return rem


def get_dim_imgs(cut_imgs):
    """
    :param cut_imgs: list of images after the cut
    :return: dimensions of images
    """
    dim = []
    for i in cut_imgs:
        n = i
        dim.append(np.asarray(i).shape)
    return dim


def get_default_overlaps(n_s_row_max, n_s_col_max):
    """
    comfortable for not defining new overlaps every time
    :param n_s_row_max: max number of rows in cut images
    :param n_s_col_max: max number of cols in cut images
    :return: our defined default overlaps
    """
    return [(n_s_row_max // 2), (n_s_col_max // 2)]


def get_slice(index, length, pad):
    """
    Given the index at which we are it returns the slice of the next img.
    (the operation is the same for width and height).
    """
    slice_ret = slice(pad * index, (pad * index) + length)
    return slice_ret


def get_edge_start_slice(rem, overlap):
    """
    :param rem: remaining dim value (W or H)
    :param overlap: overlap value
    :return: return the start point of the slice in the edge case (both in h and w)
    """
    start = (- rem - overlap)
    return start


def get_range(dim_orig_im, n_s_dim_max, pad):
    """
    Useful to give less importance to the remaining value (edge cases).
    """
    r = range(((dim_orig_im - n_s_dim_max) // pad + 1))
    return r


import numpy as np
import statistics


def reorder_cut_img(images, n_img_per_col):
    """
    :param images: list of 3d images
    :param n_img_per_col: number of images per col
    :return: a matrix of images [n_img_per_col][len(images) - n_img_per_col]
    """
    print("rebuilding preparation")
    m_img = []
    test = len(images)
    n_img_per_row = len(images) // n_img_per_col
    for r in range(n_img_per_col):
        # new row
        m_img.append([])
        for c in range(n_img_per_row):
            # we are at the image: r * n_img_per_row + c
            m_img[r].append(images[r * n_img_per_row + c])
    # now we should have each image in the right place (as in sol of the tests for cut)
    return m_img


def step1_to_rebuild(m_images, n_s_row_max, n_s_col_max, overlap_w, overlap_h, W=4172, H=1202):
    """
    :param m_images: matrix of images
    :param n_s_row_max: number of rows max for each img
    :param n_s_col_max: number of cols max for each img
    :param overlap_w:
    :param overlap_h:
    :param W: width of the original img
    :param H: height of the original img
    :return: a matrix of list of pixels as big as the original image
    """
    print("start step 1 rebuilding")
    img_step1 = [[[] for i in range(W)] for j in range(H)]
    for r in range(len(m_images)):
        for c in range(len(m_images[r])):
            c_in_step_2, r_in_step_2 = get_start_index(r, c, overlap_w, overlap_h, n_s_row_max, n_s_col_max)

            # To handle easily edge cases
            if c == len(m_images[r]) - 1:
                c_in_step_2 = W - len(m_images[r][c][0])
            if r == len(m_images) - 1:
                r_in_step_2 = H - len(m_images[r][c])

            # print("c: ", c_in_step_2, " w: ", r_in_step_2)

            for r1 in m_images[r][c]:
                tmp = c_in_step_2
                for c1 in r1:
                    img_step1[r_in_step_2][tmp].append(c1)
                    tmp += 1
                r_in_step_2 += 1
    return img_step1


def get_start_index(r, c, overlap_w, overlap_h, n_s_row_max, n_s_col_max):
    # It does not handle edge cases
    pad_h = n_s_row_max - overlap_h
    pad_w = n_s_col_max - overlap_w
    return c * pad_w, r * pad_h


def final_step_rebuild(img_step1, interpolator=lambda x: mean(x)):
    """
    :param img_step1: return image of step1_to_rebuild
    :param interpolator: function which decide how to handle the overlapped value of the pixels
    :return: starting image
    """
    print("start rebuilding")
    img_final = []
    for r in range(len(img_step1)):
        # add row
        img_final.append([])
        for c in range(len(img_step1[r])):
            param = [[] for i in range(len(img_step1[r][c][0]))]
            for p in range(len(img_step1[r][c])):
                for v in range(len(img_step1[r][c][p])):
                    param[v].append(img_step1[r][c][p][v])
            pix_to_add = []
            for l_to_int in param:
                pix_to_add.append(interpolator(l_to_int))
            # add pix
            img_final[r].append(pix_to_add)
    print("rebuilt")
    return img_final


def mean(l):
    return statistics.mean(l)


def median(l):
    return statistics.median(l)


def rebuild(images, n_img_per_col, n_s_row_max, n_s_col_max, overlap_w, overlap_h, W, H, interpolator=lambda x: mean(x)):
    """
    :param images: list of images
    :param n_img_per_col: number of images per col (regarding original img)
    :param n_s_row_max: max number of rows for single image
    :param n_s_col_max: max number of cols for single image
    :param overlap_w: overlap value on w
    :param overlap_h: overlap value on h
    :param W: W of original image in pix
    :param H: H of original image in pix
    :param interpolator: function which decide how to handle the overlapped value of the pixels
    :return: image rebuilt
    """
    m_images = reorder_cut_img(images, n_img_per_col)
    step1 = step1_to_rebuild(m_images, n_s_row_max, n_s_col_max, overlap_w, overlap_h, W, H)
    return final_step_rebuild(step1, interpolator)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='icbs',
    version='0.0.2',
    description='Simple image cut and rebuild package. It allows to decide how to cut an image in overlapping squares. After having cut the image you can rebuild it (or you can do some processing before rebuilding it)',
    long_description=long_description,
    author='Lorenzo Carpaneto',
	packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
	url="https://github.com/lolloz98/icbs",
    python_requires='>=3.6',
	install_requires=['numpy']
)
